"""
OASIS file format readers and writers

Note that OASIS references follow the same convention as `masque`,
  with this order of operations:
   1. Mirroring
   2. Rotation
   3. Scaling
   4. Offset and array expansion (no mirroring/rotation/scaling applied to offsets)

  Scaling, rotation, and mirroring apply to individual instances, not grid
   vectors or offsets.
"""
from typing import List, Any, Dict, Tuple, Callable, Union, Sequence, Iterable, Optional
import re
import io
import copy
import numpy
import base64
import struct
import logging
import pathlib
import gzip

import fatamorgana
import fatamorgana.records as fatrec
from fatamorgana.basic import PathExtensionScheme

from .utils import mangle_name, make_dose_table
from .. import Pattern, SubPattern, GridRepetition, PatternError, Label, Shape, subpattern_t
from ..shapes import Polygon, Path
from ..utils import rotation_matrix_2d, get_bit, set_bit, vector2, is_scalar, layer_t
from ..utils import remove_colinear_vertices, normalize_mirror


logger = logging.getLogger(__name__)


path_cap_map = {
                PathExtensionScheme.Flush: Path.Cap.Flush,
                PathExtensionScheme.HalfWidth: Path.Cap.Square,
                PathExtensionScheme.Arbitrary: Path.Cap.SquareCustom,
               }


def write(patterns: Union[Pattern, List[Pattern]],
          stream: io.BufferedIOBase,
          units_per_micron: int,
          layer_map: Dict[str, Union[int, Tuple[int, int]]] = None,
          modify_originals: bool = False,
          disambiguate_func: Callable[[Iterable[Pattern]], None] = None):
    """
    Write a `Pattern` or list of patterns to a OASIS file, writing patterns
     as OASIS cells, polygons as Polygon records, and subpatterns as Placement
     records. Other shape types may be converted to polygons if no equivalent
     record type exists (or is not implemented here yet). #TODO

     For each shape,
        layer is chosen to be equal to `shape.layer` if it is an int,
            or `shape.layer[0]` if it is a tuple
        datatype is chosen to be `shape.layer[1]` if available,
            otherwise `0`
        If a layer map is provided, layer strings will be converted
            automatically, and layer names will be written to the file.

    If you want pattern polygonized with non-default arguments, just call `pattern.polygonize()`
     prior to calling this function.

    Args:
        patterns: A Pattern or list of patterns to write to file.
        stream: Stream object to write to.
        units_per_micron: Written into the OASIS file, number of grid steps per micrometer.
            All distances are assumed to be an integer multiple of the grid step, and are stored as such.
        layer_map: Dictionary which translates
        modify_originals: If `True`, the original pattern is modified as part of the writing
            process. Otherwise, a copy is made and `deepunlock()`-ed.
            Default `False`.
        disambiguate_func: Function which takes a list of patterns and alters them
            to make their names valid and unique. Default is `disambiguate_pattern_names`.
    """
    if isinstance(patterns, Pattern):
        patterns = [patterns]

    if layer_map is None:
        layer_map = {}

    if disambiguate_func is None:
        disambiguate_func = disambiguate_pattern_names

    if not modify_originals:
        patterns = [p.deepunlock() for p in copy.deepcopy(patterns)]

    # Create library
    lib = fatamorgana.OasisLayout(unit, validation=None)

    for name, layer_num in layer_map.items():
        layer, data_type = _mlayer2oas(layer_num)
        lib.layer_names.append(                     #TODO figure out how to deal with text layers
                LayerName(nstring=name,
                          layer_interval=(layer, layer),
                          type_interval=(data_type, data_type),
                          is_textlayer=False))

    def layer2oas(layer: layer_t) -> Tuple[int, int]:
        layer_num = layer_map[layer] if isinstance(layer, str) else layer
        return _mlayer2oas(layer_num)

    # Get a dict of id(pattern) -> pattern
    patterns_by_id = {id(pattern): pattern for pattern in patterns}
    for pattern in patterns:
        for i, p in pattern.referenced_patterns_by_id().items():
            patterns_by_id[i] = p

    disambiguate_func(patterns_by_id.values())

    # Now create a structure for each pattern
    for pat in patterns_by_id.values():
        structure = fatamorgana.Cell(name=NString(pat.name))
        lib.cells.append(structure)

        structure.geometry += _shapes_to_elements(pat.shapes, layer2oas)
        structure.geometry += _labels_to_texts(pat.labels, layer2oas)
        structure.placements += _subpatterns_to_refs(pat.subpatterns)

    lib.write(stream)
    return


def writefile(patterns: Union[List[Pattern], Pattern],
              filename: Union[str, pathlib.Path],
              *args,
              **kwargs,
              ):
    """
    Wrapper for `oasis.write()` that takes a filename or path instead of a stream.

    Will automatically compress the file if it has a .gz suffix.

    Args:
        patterns: `Pattern` or list of patterns to save
        filename: Filename to save to.
        *args: passed to `oasis.write`
        **kwargs: passed to `oasis.write`
    """
    path = pathlib.Path(filename)
    if path.suffix == '.gz':
        open_func: Callable = gzip.open
    else:
        open_func = open

    with io.BufferedWriter(open_func(path, mode='wb')) as stream:
        results = write(patterns, stream, *args, **kwargs)
    return results


def readfile(filename: Union[str, pathlib.Path],
             *args,
             **kwargs,
             ) -> Tuple[Dict[str, Pattern], Dict[str, Any]]:
    """
    Wrapper for `oasis.read()` that takes a filename or path instead of a stream.

    Will automatically decompress files with a .gz suffix.

    Args:
        filename: Filename to save to.
        *args: passed to `oasis.read`
        **kwargs: passed to `oasis.read`
    """
    path = pathlib.Path(filename)
    if path.suffix == '.gz':
        open_func: Callable = gzip.open
    else:
        open_func = open

    with io.BufferedReader(open_func(path, mode='rb')) as stream:
        results = read(stream, *args, **kwargs)
    return results


def read(stream: io.BufferedIOBase,
         clean_vertices: bool = True,
         ) -> Tuple[Dict[str, Pattern], Dict[str, Any]]:
    """
    Read a OASIS file and translate it into a dict of Pattern objects. OASIS cells are
     translated into Pattern objects; Polygons are translated into polygons, and Placements
     are translated into SubPattern or GridRepetition objects.

    Additional library info is returned in a dict, containing:
      'units_per_micrometer': number of database units per micrometer (all values are in database units)

    Args:
        stream: Stream to read from.
        clean_vertices: If `True`, remove any redundant vertices when loading polygons.
            The cleaning process removes any polygons with zero area or <3 vertices.
            Default `True`.

    Returns:
        - Dict of pattern_name:Patterns generated from GDSII structures
        - Dict of GDSII library info
    """

    lib = fatamorgana.OASISLayout.read(stream)

    library_info = {'units_per_micrometer': lib.unit,
                    }

    patterns = []
    for cell in lib.cells:
        pat = Pattern(name=cell.name.string)
        for element in cell.geometry:
            if element.repetition is not None:
                raise PatternError('masque OASIS reader does not implement repetitions for shapes yet')

            # Switch based on element type:
            if isinstance(element, fatrec.Polygon):
                args = {'vertices': element.point_list,
                        'layer':  (element.layer, element.data_type)
                        'offset': (element.x, element.y),
                       }
                poly = Polygon(**args)

                if clean_vertices:
                    try:
                        poly.clean_vertices()
                    except PatternError:
                        continue

                pat.shapes.append(poly)

            if isinstance(element, fatrec.Path):
                cap_start = path_cap_map[element.extension_start[0]]
                cap_end   = path_cap_map[element.extension_end[0]]
                if cap_start != cap_end:
                    raise Exception('masque does not support multiple cap types on a single path.')      #TODO handle multiple cap types
                cap = cap_start

                args = {'vertices': element.point_list,
                        'layer':  (element.layer, element.data_type)
                        'offset': (element.x, element.y),
                        'width': element.half_width * 2,
                        'cap': cap,
                       }

                if cap == Path.Cap.SquareCustom:
                    args['cap_extensions'] = numpy.array((element.extension_start[1],
                                                          element.extension_end[1]))
                path = Path(**args)

                if clean_vertices:
                    try:
                        path.clean_vertices()
                    except PatternError as err:
                        continue

                pat.shapes.append(path)

            elif isinstance(element, fatrec.Text):
                args = {'layer':  (element.layer, element.data_type)
                        'offset': (element.x, element.y),
                        'string': str(element.string),
                       }
                pat.labels.append(Label(**args))

        for placement in cell.placements:
            pat.subpattterns.append += _placement_to_subpats(placement)

        patterns.append(pat)

    # Create a dict of {pattern.name: pattern, ...}, then fix up all subpattern.pattern entries
    #  according to the subpattern.identifier (which is deleted after use).
    patterns_dict = dict(((p.name, p) for p in patterns))
    for p in patterns_dict.values():
        for sp in p.subpatterns:
            sp.pattern = patterns_dict[sp.identifier[0]]
            del sp.identifier

    return patterns_dict, library_info


def _mlayer2oas(mlayer: layer_t) -> Tuple[int, int]:
    """ Helper to turn a layer tuple-or-int into a layer and datatype"""
    if isinstance(mlayer, int):
        layer = mlayer
        data_type = 0
    elif isinstance(mlayer, tuple):
        layer = mlayer[0]
        if len(mlayer) > 1:
            data_type = mlayer[1]
        else:
            data_type = 0
    else:
        raise PatternError(f'Invalid layer for OASIS: {layer}. Note that OASIS layers cannot be strings.')  #TODO allow string layers using layer map def
    return layer, data_type


def _placement_to_subpats(placement: fatrec.Placement) -> List[subpattern_t]:
    """
    Helper function to create a SubPattern from a placment. Sets subpat.pattern to None
     and sets the instance .identifier to (struct_name,).
    """
    xy = numpy.array((placement.x, placement.y))
    kwargs = {
       'pattern': None,
       'mirrored': (placement.flip, False),
       'rotation': float(placement.angle * pi/180)
       'scale': placement.magnification,
       'identifier': (placement.name,),
       }

    rep = placement.repetition
    if isinstance(rep, fatamorgana.GridRepetition):
        subpat = GridRepetition(a_vector=rep.a_vector,
                                b_vector=rep.b_vector,
                                a_count=rep.a_count,
                                b_count=rep.b_count,
                                offset=xy,
                                **kwargs)
        subpats = [subpat]
    elif isinstance(rep, fatamorgana.ArbitraryRepetition):
        subpats = []
        for rep_offset in numpy.cumsum(numpy.column_stack((rep.x_displacements,
                                                           rep.y_displacements))):
            subpats.append(SubPattern(offset=xy + rep_offset, **kwargs))
    elif rep is None
        subpats = [SubPattern(offset=xy + rep_offset, **kwargs)]
    return subpats


def _subpatterns_to_refs(subpatterns: List[subpattern_t]
                        ) -> List[fatrec.Placement]]:
    refs = []
    for subpat in subpatterns:
        if subpat.pattern is None:
            continue

        # Note: OASIS mirrors first and rotates second
        mirror_across_x, extra_angle = normalize_mirror(subpat.mirrored)
        xy = numpy.round(subpat.offset).astype(int)
        args = {
            'x': xy[0],
            'y': xy[1],
            }

        if isinstance(subpat, GridRepetition):
            kwargs['rep'] = fatamorgana.GridRepetition(
                              a_vector=numpy.round(subpat.a_vector).astype(int),
                              b_vector=numpy.round(subpat.b_vector).astype(int),
                              a_count=numpy.round(subpat.a_count).astype(int),
                              b_count=numpy.round(subpat.b_count).astype(int))

        ref = fatrec.Placement(
                name=subpat.pattern.name,
                flip=mirror_across_x,
                angle=((subpat.rotation + extra_angle) * 180 / numpy.pi) % 360,
                magnification=subpat.scale,
                **kwargs)

        refs.append(ref)
    return refs


def _shapes_to_elements(shapes: List[Shape],
                        layer2oas: Callable[[layer_t], Tuple[int, int]],
                        polygonize_paths: bool = False,
                       ) -> List[Union[fatrec.Polygon, fatrec.Path]]:
    # Add a Polygon record for each shape, and Path elements if necessary
    elements: List[Union[fatrec.Polygon, fatrec.Path]] = []
    for shape in shapes:
        layer, data_type = layer2oas(shape.layer)
        if isinstance(shape, Path) and not polygonize_paths:
            offset = numpy.round(shape.offset).astype(int)
            points = numpy.round(shape.vertices).astype(int)
            half_width = numpy.round(shape.width / 2).astype(int)
            path_type = next(k for k, v in path_cap_map.items() if v == shape.cap)    #reverse lookup
            path = fatrec.Path(layer=layer,
                               data_type=data_type,
                               point_list=points,
                               half_width=half_width,
                               x=offset[0],
                               y=offset[1],
                               extension_start=path_type,       #TODO implement multiple cap types?
                               extension_end=path_type,
                               )
            elements.append(path)
        else:
            for polygon in shape.to_polygons():
                points = numpy.round(polygon.vertices).astype(int)
                offset = numpy.round(polygon.offset).astype(int)
                elements.append(fatrec.Polygon(layer=layer,
                                               data_type=data_type,
                                               x=offset[0],
                                               y=offset[1],
                                               point_list=point_list))
    return elements


def _labels_to_texts(labels: List[Label],
                     layer2oas: Callable[[layer_t], Tuple[int, int]],
                     ) -> List[fatrec.Text]:
    texts = []
    for label in labels:
        layer, text_type = layer2oas(label.layer)
        xy = numpy.round(label.offset).astype(int)
        texts.append(fatrec.Text(layer=layer,
                                 text_type=text_type,
                                 x=xy[0],
                                 y=xy[1],
                                 string=string))
    return texts


def disambiguate_pattern_names(patterns,
                               dup_warn_filter: Callable[[str,], bool] = None,      # If returns False, don't warn about this name
                               ):
    used_names = []
    for pat in patterns:
        sanitized_name = re.compile('[^A-Za-z0-9_\?\$]').sub('_', pat.name)

        i = 0
        suffixed_name = sanitized_name
        while suffixed_name in used_names or suffixed_name == '':
            suffix = base64.b64encode(struct.pack('>Q', i), b'$?').decode('ASCII')

            suffixed_name = sanitized_name + '$' + suffix[:-1].lstrip('A')
            i += 1

        if sanitized_name == '':
            logger.warning('Empty pattern name saved as "{}"'.format(suffixed_name))
        elif suffixed_name != sanitized_name:
            if dup_warn_filter is None or dup_warn_filter(pat.name):
                logger.warning('Pattern name "{}" ({}) appears multiple times;\n renaming to "{}"'.format(
                                pat.name, sanitized_name, suffixed_name))

        encoded_name = suffixed_name.encode('ASCII')
        if len(encoded_name) == 0:
            # Should never happen since zero-length names are replaced
            raise PatternError('Zero-length name after sanitize+encode,\n originally "{}"'.format(pat.name))

        pat.name = encoded_name
        used_names.append(suffixed_name)
